#!/usr/bin/python3
import os
from sklearn.metrics import precision_score
from sklearn.model_selection import cross_val_predict
from sklearn import model_selection
from sklearn import metrics
from sklearn.datasets import make_classification
from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score
from sklearn.metrics import mean_squared_error

import numpy as np
import statistics
import collections
from sklearn import datasets, linear_model
import numpy.polynomial.polynomial as poly
from sklearn.pipeline import make_pipeline
from sklearn.base import BaseEstimator, TransformerMixin

from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn import utils
from sklearn.utils import shuffle
import time


def fit_string_data(dataset, headlist):
    from sklearn.preprocessing import LabelEncoder
    lb = LabelEncoder()
    for each in headlist:
        dataset[each] = lb.fit_transform(dataset[each])
    return dataset

def get_CV_forefficiency(classicclassifier, X, y,nfold):
    cvaccuracylist = []
    for i in range(0,nfold,1):
        X_train, X_valid, y_train, y_valid = model_selection.train_test_split(X, y, test_size=1.0/nfold)
        classicclassifier.fit(X_train,y_train)
        cvaccuracy = accuracy_score(y_valid,classicclassifier.predict(X_valid))
        cvaccuracylist.append(cvaccuracy)
    return statistics.mean(cvaccuracylist)


def get_PV_classic_forefficiency(classicclassifier, X, y):
    noisedegreelist = [0,0.2]
    label_list = list(set(list(y)))

    dic_label_num = {} # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y == label).sum()

    model = classicclassifier

    trainingacculist = [] # list to put perturbed training accuracy
    model.fit(X, y)
    trainaccuracy_original = accuracy_score(y, model.predict(X))
    trainingacculist.append(trainaccuracy_original)
    noisedegree = noisedegreelist[1]


    Y_changed = np.copy(y)
    dic_label_newnum = {} # record the number of perturbed samples for each class

    for label in label_list:
        dic_label_newnum[label] = 0 #initialize the dic; nothing is perturbed at the beginning

    for i in range(0, Y_changed.size):
        cnnn = 0
        for label in label_list:
            # perturb the label only if the perturbed labels in this class are fewer than required,
            if dic_label_newnum[label_list[cnnn]] < float(dic_label_num[label_list[cnnn]]) * noisedegree and Y_changed[i] == label_list[cnnn]:
                try:
                    # replace the current label with its right neighbour label in label_list
                    Y_changed[i] = label_list[cnnn+1]
                    dic_label_newnum[label_list[cnnn]] += 1
                except:
                    # if the label is the last in the label_list, replace it with the first element in label_list
                    Y_changed[i] = label_list[0]
                    dic_label_newnum[label_list[cnnn]] += 1
                continue
            cnnn += 1


    model.fit(X, Y_changed) # retrain the model with perturbed labels
    y_changed_predictions = model.predict(X)
    trainaccuracy_perturbed = accuracy_score(Y_changed, y_changed_predictions)
    trainaccuracy_withoriginal = accuracy_score(y, y_changed_predictions)
    trainingacculist.append(trainaccuracy_perturbed)


    Ytest = trainingacculist


    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    pv = -b # mirror PV by one, so that PV increases up to 1, then worsen afterwards
    pv = 0.6 * trainaccuracy_withoriginal + 0.2 * pv + 0.2


    return pv


def get_PV_classic(classicclassifier, X, y):
    '''
    Description
        This function calculates the PV (model fit) evaluation result of a classic classifier and a training data set.

    Parameters
        classifier: the learner adopted to train the data; type: sklearn classifier
        X: the training data instances (without labels); type: numpy.ndarray
        y: training data labels; type: numpy.ndarray

    Returns
        PV: the degree of fit between classifier and data; type: float

    Examples
        from sklearn.tree import DecisionTreeClassifier
        from sklearn import datasets
        iris = datasets.load_iris()
        X = iris.data
        y = iris.target
        dt = DecisionTreeClassifier()
        pv = lib.get_PV_classic(dt,X,y)
    '''

    #X, y = shuffle(X, y)
    noisedegreelist = [0,0.2]
    label_list = list(set(list(y)))

    dic_label_num = {} # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y == label).sum()

    model = classicclassifier

    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0
    while (cnt < len(noisedegreelist)):
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in label_list:
            dic_label_newnum[label] = 0 #initialize the dic; nothing is perturbed at the beginning

        for i in range(0, Y_changed.size):
            cnnn = 0
            for label in label_list:
                # perturb the label only if the perturbed labels in this class are fewer than required,
                if Y_changed[i] == label_list[cnnn] and dic_label_newnum[label_list[cnnn]] < float(dic_label_num[label_list[cnnn]]) * noisedegree:
                    try:
                        # replace the current label with its right neighbour label in label_list
                        Y_changed[i] = label_list[cnnn+1]
                        dic_label_newnum[label_list[cnnn]] += 1
                    except:
                        # if the label is the last in the label_list, replace it with the first element in label_list
                        Y_changed[i] = label_list[0]
                        dic_label_newnum[label_list[cnnn]] += 1
                    continue
                cnnn += 1


        model.fit(X, Y_changed) # retrain the model with perturbed labels
        y_changed_predictions = model.predict(X)
        trainaccuracy_perturbed = accuracy_score(Y_changed, y_changed_predictions)
        trainaccuracy_withoriginal = accuracy_score(y, y_changed_predictions)
        trainingacculist.append(trainaccuracy_perturbed)
        cnt+=1

    Ytest = trainingacculist

    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    pv = -b # mirror PV by one, so that PV increases up to 1, then worsen afterwards
    # pv = trainaccuracy_perturbed
    pv = 0.6 * trainaccuracy_withoriginal + 0.2 * pv + 0.2
    # print(trainingacculist)
    #pv = -b

    # print('Perturbation results:')
    # for i in np.arange(0, 2, 1):
    #     print('Label noise degree: ' + str(round(noisedegreelist[i], 2)) + '  Training accuracy: ' + str(
    #         round(trainingacculist[i],2)))
    #
    # print('PV: '+str(round(pv,2)))
    # print('----------')

    return pv

def get_PV_classic_withdifferentnoisedegree(classicclassifier, X, y,noisedegreepara):
    '''
    Description
        This function calculates the PV (model fit) evaluation result of a classic classifier and a training data set.

    Parameters
        classifier: the learner adopted to train the data; type: sklearn classifier
        X: the training data instances (without labels); type: numpy.ndarray
        y: training data labels; type: numpy.ndarray

    Returns
        PV: the degree of fit between classifier and data; type: float

    Examples
        from sklearn.tree import DecisionTreeClassifier
        from sklearn import datasets
        iris = datasets.load_iris()
        X = iris.data
        y = iris.target
        dt = DecisionTreeClassifier()
        pv = lib.get_PV_classic(dt,X,y)
    '''

    # X, y = shuffle(X, y)
    noisedegreelist = [0, noisedegreepara]
    label_list = list(set(list(y)))

    dic_label_num = {} # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y == label).sum()

    model = classicclassifier

    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0
    while (cnt < len(noisedegreelist)):
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in label_list:
            dic_label_newnum[label] = 0 #initialize the dic; nothing is perturbed at the beginning

        for i in range(0, Y_changed.size):
            cnnn = 0
            for label in label_list:
                # perturb the label only if the perturbed labels in this class are fewer than required,
                if Y_changed[i] == label_list[cnnn] and dic_label_newnum[label_list[cnnn]] < float(dic_label_num[label_list[cnnn]]) * noisedegree:
                    try:
                        # replace the current label with its right neighbour label in label_list
                        Y_changed[i] = label_list[cnnn+1]
                        dic_label_newnum[label_list[cnnn]] += 1
                    except:
                        # if the label is the last in the label_list, replace it with the first element in label_list
                        Y_changed[i] = label_list[0]
                        dic_label_newnum[label_list[cnnn]] += 1
                    continue
                cnnn += 1


        model.fit(X, Y_changed) # retrain the model with perturbed labels
        y_changed_predictions = model.predict(X)
        trainaccuracy_perturbed = accuracy_score(Y_changed, y_changed_predictions)
        trainaccuracy_withoriginal = accuracy_score(y, y_changed_predictions)
        trainingacculist.append(trainaccuracy_perturbed)
        cnt+=1

    Ytest = trainingacculist

    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    pv = -b # mirror PV by one, so that PV increases up to 1, then worsen afterwards
    # pv = trainaccuracy_perturbed
    pv = (1-2*noisedegree)*trainaccuracy_withoriginal+noisedegree*pv+noisedegree
    # print(trainingacculist)
    #pv = -b

    # print('Perturbation results:')
    # for i in np.arange(0, 2, 1):
    #     print('Label noise degree: ' + str(round(noisedegreelist[i], 2)) + '  Training accuracy: ' + str(
    #         round(trainingacculist[i],2)))
    #
    # print('PV: '+str(round(pv,2)))
    # print('----------')

    return pv

def get_PV_classic_mse(classicclassifier, X, y):
    '''
    Description
        This function calculates the PV (model fit) evaluation result of a classic classifier and a training data set.

    Parameters
        classifier: the learner adopted to train the data; type: sklearn classifier
        X: the training data instances (without labels); type: numpy.ndarray
        y: training data labels; type: numpy.ndarray

    Returns
        PV: the degree of fit between classifier and data; type: float

    Examples
        from sklearn.tree import DecisionTreeClassifier
        from sklearn import datasets
        iris = datasets.load_iris()
        X = iris.data
        y = iris.target
        dt = DecisionTreeClassifier()
        pv = lib.get_PV_classic(dt,X,y)
    '''

    X, y = shuffle(X, y)
    noisedegreelist = [0,0.2]
    label_list = list(set(list(y)))

    dic_label_num = {} # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y == label).sum()

    model = classicclassifier

    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0
    while (cnt < len(noisedegreelist)):
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in label_list:
            dic_label_newnum[label] = 0 #initialize the dic; nothing is perturbed at the beginning

        for i in range(0, Y_changed.size):
            cnnn = 0
            for label in label_list:
                # perturb the label only if the perturbed labels in this class are fewer than required,
                if Y_changed[i] == label_list[cnnn] and dic_label_newnum[label_list[cnnn]] < float(dic_label_num[label_list[cnnn]]) * noisedegree:
                    try:
                        # replace the current label with its right neighbour label in label_list
                        Y_changed[i] = label_list[cnnn+1]
                        dic_label_newnum[label_list[cnnn]] += 1
                    except:
                        # if the label is the last in the label_list, replace it with the first element in label_list
                        Y_changed[i] = label_list[0]
                        dic_label_newnum[label_list[cnnn]] += 1
                    continue
                cnnn += 1


        model.fit(X, Y_changed) # retrain the model with perturbed labels
        y_changed_predictions = model.predict(X)
        trainaccuracy_perturbed = 1-mean_squared_error(Y_changed, y_changed_predictions)
        trainaccuracy_withoriginal = 1- mean_squared_error(y, y_changed_predictions)
        trainingacculist.append(trainaccuracy_perturbed)
        cnt+=1

    Ytest = trainingacculist

    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    pv = -b # mirror PV by one, so that PV increases up to 1, then worsen afterwards
    # pv = trainaccuracy_perturbed
    pv = 0.6 * trainaccuracy_withoriginal + 0.2 * pv + 0.2
    # print(trainingacculist)
    #pv = -b

    # print('Perturbation results:')
    # for i in np.arange(0, 2, 1):
    #     print('Label noise degree: ' + str(round(noisedegreelist[i], 2)) + '  Training accuracy: ' + str(
    #         round(trainingacculist[i],2)))
    #
    # print('PV: '+str(round(pv,2)))
    # print('----------')

    return pv

def unique_rows(a):
    a = np.ascontiguousarray(a)
    unique_a = np.unique(a.view([('', a.dtype)]*a.shape[1]))
    return unique_a.view(a.dtype).reshape((unique_a.shape[0], a.shape[1]))



def get_PV_deep(create_deepclassifier, X, y,epoch):
    '''
    Description
        This function calculates the PV (model fit) evaluation result of a deep learning classifier and a training data set.

    Parameters
        create_deepclassifier: function to create a classifier;
        X: the training data instances (without labels); type: numpy.ndarray
        y: training data labels; type: numpy.ndarray
        epoch: epoch used to train the model

    Returns
        PV: the degree of fit between classifier and data; type: float
    '''
    noisedegreelist = np.arange(0.0,0.31,0.1) # label noise sequence [0,0.1,0.2,0.3]
    print(y)
    labeldic = {}
    label_list = list(unique_rows(y))
    labelid = 0
    for each in label_list:
        labeldic[labelid] = each
        labelid += 1


    dic_label_num = {} # get the number of instances for each class

    for label in labeldic:
        dic_label_num[label] = 0
        for eachitem in y:
            if np.array_equal(labeldic[label],eachitem):
                dic_label_num[label]+= 1


    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0
    while (cnt < len(noisedegreelist)):
        model = create_deepclassifier()
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in labeldic:
            dic_label_newnum[label] = 0 # initialize the dic; nothing is perturbed at the beginning

        for i in range(0, y.shape[0]):
            for eachelement in labeldic:
                if np.array_equal(Y_changed[i], labeldic[eachelement]) and (dic_label_newnum[eachelement] < float(dic_label_num[eachelement]) * noisedegree):
                    try:
                        Y_changed[i] = labeldic[eachelement + 1]
                        dic_label_newnum[eachelement] += 1
                    except:
                        Y_changed[i] = labeldic[0]
                        dic_label_newnum[eachelement] += 1
                    break
                else:
                    continue

        model.fit(X, Y_changed,batch_size=128,epochs=epoch, verbose=1) #retrain the model with perturbed labels

        trainaccuracy_perturbed = model.evaluate(X,Y_changed,verbose=0)[1]
        trainingacculist.append(trainaccuracy_perturbed)
        cnt+=1

    Ytest = trainingacculist

    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    pv = 1-abs(1-abs(b)) # mirror PV by one, so that PV increases up to 1, then worsen afterwards

    print('Perturbation results:')
    for i in np.arange(0,4,1):
        print('Label noise degree: '+str(round(noisedegreelist[i],2))+'  Training accuracy: '+str(round(trainingacculist[i],3)))
    print('PV: ' + str(round(pv, 2)))

    return pv


def get_allmetrics_classic(classicclassifier, X_train, y_train, X_hold, y_hold):
    '''
    Description
        This function calculates the values of all metrics
        (training accuracy, test accuracy, 3-fold CV accuracy, and PV)
        of a classic classifier.

    Parameters
        classifier: the learner adopted to train the data; type: sklearn classifier
        X_train: the training data instances (without labels); type: numpy.ndarray
        y_train: training data labels; type: numpy.ndarray
        X_hold: the hold-out data instances (without labels); type: numpy.ndarray
        y_hold: the hold-out data labels; type: numpy.ndarray

    Returns
        dic_metric_value: the dic of metric values
        key (string) of the dic:
            pv: PV
            cv: CV accuracy
            train: training accuracy
            test: test accuracy
    '''
    noisedegreelist = [0,0.2]
    label_list = list(set(list(y_train)))

    dic_label_num = {} # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y_train == label).sum()


    model = classicclassifier

    # get 3-fold CV accuracy

    cvresults = model_selection.cross_val_score(model, X_train, y_train, cv=3)
    cvaccuracy = cvresults.mean()

    # get test accuracy
    m = model.fit(X_train, y_train)
    y_test_predictions = m.predict(X_hold)
    testaccuracy = accuracy_score(y_hold, y_test_predictions)


    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0
    while (cnt < len(noisedegreelist)):
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y_train)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in label_list:
            dic_label_newnum[label] = 0 #initialize the dic; nothing is perturbed at the beginning

        for i in range(0, Y_changed.size):
            cnnn = 0
            for label in label_list:
                # perturb the label only if the perturbed labels in this class are fewer than required,
                if Y_changed[i] == label_list[cnnn] and dic_label_newnum[label_list[cnnn]] < float(dic_label_num[label_list[cnnn]]) * noisedegree:
                    try:
                        # replace the current label with its right neighbour label in label_list
                        Y_changed[i] = label_list[cnnn+1]
                        dic_label_newnum[label_list[cnnn]] += 1
                    except:
                        # if the label is the last in the label_list, replace it with the first element in label_list
                        Y_changed[i] = label_list[0]
                        dic_label_newnum[label_list[cnnn]] += 1
                    continue
                cnnn += 1



        model.fit(X_train, Y_changed) # retrain the model with perturbed labels
        y_changed_predictions = model.predict(X_train)
        trainaccuracy_perturbed = accuracy_score(Y_changed, y_changed_predictions)
        trainaccuracy_withoriginal = accuracy_score(y_train, y_changed_predictions)
        trainingacculist.append(trainaccuracy_perturbed)
        cnt+=1

    Ytest = trainingacculist

    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    # pv = 1-abs(1-abs(b)) # mirror PV by one, so that PV increases up to 1, then worsen afterwards
    pv = -b
    # pv = trainaccuracy_perturbed
    # pv = 0.6*trainaccuracy_withoriginal+0.2*pv+0.2
    pv = 0.6 * trainaccuracy_withoriginal + (trainingacculist[0]-trainingacculist[1]) + 0.2

    if pv>1:
        pv = 1

    # print('PV: '+str(round(pv,2)))
    # print('training accuracy: '+str(round(trainingacculist[0],2)))
    # print('CV accuracy: '+str(round(cvaccuracy,2)))
    # print('test accuracy: '+str(round(testaccuracy,2)))
    # print('----------')


    # return dic
    dic_metric_value = {}
    dic_metric_value['pv'] = pv
    dic_metric_value['train'] = trainingacculist[0]
    dic_metric_value['test'] = testaccuracy
    dic_metric_value['cv'] = cvaccuracy

    return dic_metric_value


def get_allmetrics_classic_withdifferentnoisedegree(classicclassifier, X_train, y_train, X_hold, y_hold,noisedegree):
    '''
    Description
        This function calculates the values of all metrics
        (training accuracy, test accuracy, 3-fold CV accuracy, and PV)
        of a classic classifier.

    Parameters
        classifier: the learner adopted to train the data; type: sklearn classifier
        X_train: the training data instances (without labels); type: numpy.ndarray
        y_train: training data labels; type: numpy.ndarray
        X_hold: the hold-out data instances (without labels); type: numpy.ndarray
        y_hold: the hold-out data labels; type: numpy.ndarray

    Returns
        dic_metric_value: the dic of metric values
        key (string) of the dic:
            pv: PV
            cv: CV accuracy
            train: training accuracy
            test: test accuracy
    '''
    noisedegreelist = [0,noisedegree]
    label_list = list(set(list(y_train)))

    dic_label_num = {} # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y_train == label).sum()

    model = classicclassifier

    # get 3-fold CV accuracy
    cvresults = model_selection.cross_val_score(model, X_train, y_train, cv=3)
    cvaccuracy = cvresults.mean()

    # get test accuracy
    m = model.fit(X_train, y_train)
    y_test_predictions = m.predict(X_hold)
    testaccuracy = accuracy_score(y_hold, y_test_predictions)


    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0
    while (cnt < len(noisedegreelist)):
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y_train)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in label_list:
            dic_label_newnum[label] = 0 #initialize the dic; nothing is perturbed at the beginning

        for i in range(0, Y_changed.size):
            cnnn = 0
            for label in label_list:
                # perturb the label only if the perturbed labels in this class are fewer than required,
                if Y_changed[i] == label_list[cnnn] and dic_label_newnum[label_list[cnnn]] < float(dic_label_num[label_list[cnnn]]) * noisedegree:
                    try:
                        # replace the current label with its right neighbour label in label_list
                        Y_changed[i] = label_list[cnnn+1]
                        dic_label_newnum[label_list[cnnn]] += 1
                    except:
                        # if the label is the last in the label_list, replace it with the first element in label_list
                        Y_changed[i] = label_list[0]
                        dic_label_newnum[label_list[cnnn]] += 1
                    continue
                cnnn += 1



        model.fit(X_train, Y_changed) # retrain the model with perturbed labels
        y_changed_predictions = model.predict(X_train)
        trainaccuracy_perturbed = accuracy_score(Y_changed, y_changed_predictions)
        trainaccuracy_withoriginal = accuracy_score(y_train, y_changed_predictions)
        trainingacculist.append(trainaccuracy_perturbed)
        cnt+=1

    Ytest = trainingacculist

    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    # pv = 1-abs(1-abs(b)) # mirror PV by one, so that PV increases up to 1, then worsen afterwards
    pv = -b
    # pv = trainaccuracy_perturbed
    pv = (1-2*noisedegree)*trainaccuracy_withoriginal+noisedegree*pv+noisedegree



    print('PV: '+str(round(pv,2)))
    print('training accuracy: '+str(round(trainingacculist[0],2)))
    print('CV accuracy: '+str(round(cvaccuracy,2)))
    print('test accuracy: '+str(round(testaccuracy,2)))
    print('----------')


    # return dic
    dic_metric_value = {}
    dic_metric_value['pv'] = pv
    dic_metric_value['train'] = trainingacculist[0]
    dic_metric_value['test'] = testaccuracy
    dic_metric_value['cv'] = cvaccuracy

    return dic_metric_value


def get_allmetrics_classic_linearproof(classicclassifier, X_train, y_train, X_hold, y_hold,resultpath, depth):
    '''
    Description
        This function calculates the values of all metrics
        (training accuracy, test accuracy, 3-fold CV accuracy, and PV)
        of a classic classifier.

    Parameters
        classifier: the learner adopted to train the data; type: sklearn classifier
        X_train: the training data instances (without labels); type: numpy.ndarray
        y_train: training data labels; type: numpy.ndarray
        X_hold: the hold-out data instances (without labels); type: numpy.ndarray
        y_hold: the hold-out data labels; type: numpy.ndarray

    Returns
        dic_metric_value: the dic of metric values
        key (string) of the dic:
            pv: PV
            cv: CV accuracy
            train: training accuracy
            test: test accuracy
    '''
    noisedegreelist = np.arange(0.0,0.51,0.02) # label noise sequence [0,0.1,0.2,0.3]
    label_list = list(set(list(y_train)))

    writefile = open(resultpath,'a')

    dic_label_num = {} # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y_train == label).sum()

    model = classicclassifier

    # get 3-fold CV accuracy
    cvresults = model_selection.cross_val_score(model, X_train, y_train, cv=3)
    cvaccuracy = cvresults.mean()

    # get test accuracy
    m = model.fit(X_train, y_train)
    y_test_predictions = m.predict(X_hold)
    testaccuracy = accuracy_score(y_hold, y_test_predictions)


    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0
    while (cnt < len(noisedegreelist)):
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y_train)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in label_list:
            dic_label_newnum[label] = 0 #initialize the dic; nothing is perturbed at the beginning

        for i in range(0, Y_changed.size):
            cnnn = 0
            for label in label_list:
                # perturb the label only if the perturbed labels in this class are fewer than required,
                if Y_changed[i] == label_list[cnnn] and dic_label_newnum[label_list[cnnn]] < float(dic_label_num[label_list[cnnn]]) * noisedegree:
                    try:
                        # replace the current label with its right neighbour label in label_list
                        Y_changed[i] = label_list[cnnn+1]
                        dic_label_newnum[label_list[cnnn]] += 1
                    except:
                        # if the label is the last in the label_list, replace it with the first element in label_list
                        Y_changed[i] = label_list[0]
                        dic_label_newnum[label_list[cnnn]] += 1
                    continue
                cnnn += 1



        model.fit(X_train, Y_changed) # retrain the model with perturbed labels
        y_changed_predictions = model.predict(X_train)
        trainaccuracy_perturbed = accuracy_score(Y_changed, y_changed_predictions)
        trainingacculist.append(trainaccuracy_perturbed)
        writefile.write(str(depth) + ',' + str(noisedegree) + ',' + str(trainaccuracy_perturbed) + '\n')
        cnt+=1

    writefile.close()

    Ytest = trainingacculist

    Xtest = noisedegreelist
    m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    pv = 1-abs(1-abs(b)) # mirror PV by one, so that PV increases up to 1, then worsen afterwards



    print('PV: '+str(round(pv,2)))
    print('training accuracy: '+str(round(trainingacculist[0],2)))
    print('CV accuracy: '+str(round(cvaccuracy,2)))
    print('test accuracy: '+str(round(testaccuracy,2)))
    print('----------')


    # return dic
    dic_metric_value = {}
    dic_metric_value['pv'] = pv
    dic_metric_value['train'] = trainingacculist[0]
    dic_metric_value['test'] = testaccuracy
    dic_metric_value['cv'] = cvaccuracy

    return dic_metric_value


def get_PV_allmetrics_deep(create_deepclassifier, X_train, y_train, X_hold, y_hold,epoch,dropout):
    '''
    Description
        This function calculates the values of all metrics
        (training accuracy, test accuracy, validation accuracy, and PV).

    Parameters
        create_deepclassifier: the function used to initilise classifier
        X_train: the training data instances (without labels); type: numpy.ndarray
        y_train: training data labels; type: numpy.ndarray
        X_hold: the hold-out data instances (without labels); type: numpy.ndarray
        y_hold: the hold-out data labels; type: numpy.ndarray
        epoch: epochs
        dropout: dropout rate

    Returns
        dic_metric_value: the dic of metric values
        key (string) of the dic:
            pv: PV
            cv: CV accuracy
            train: training accuracy
            test: test accuracy
    '''
    noisedegreelist = [0,0.2] # label noise sequence [0,0.1,0.2,0.3]

    # get validation accuracy
    x_train2, x_valid, y_train2, y_valid = model_selection.train_test_split(X_train, y_train, test_size=0.4)
    if epoch <20:
        model = create_deepclassifier(dropout) # create model for mnist
    else:
        model = create_deepclassifier(dropout,X_train) # create model for cifar
    model.fit(x_train2, y_train2, batch_size=128, epochs=epoch, verbose=1)  # retrain the model with perturbed labels
    cvaccuracy = model.evaluate(x_valid, y_valid, verbose=0)[1]


    labeldic = {}
    label_list = list(unique_rows(y_train))
    labelid = 0
    for each in label_list:
        labeldic[labelid] = each
        labelid += 1


    dic_label_num = {} # get the number of instances for each class

    for label in labeldic:
        dic_label_num[label] = 0
        for eachitem in y_train:
            if np.array_equal(labeldic[label],eachitem):
                dic_label_num[label]+= 1


    trainingacculist = [] # list to put perturbed training accuracy

    cnt = 0

    testflag = False
    testaccuracy = 0

    while (cnt < len(noisedegreelist)):
        if epoch < 20:
            model = create_deepclassifier(dropout) # create model for mnist
        else:
            model = create_deepclassifier(dropout, X_train) # create model for cifar
        noisedegree = noisedegreelist[cnt]
        Y_changed = np.copy(y_train)
        dic_label_newnum = {} # record the number of perturbed samples for each class

        for label in labeldic:
            dic_label_newnum[label] = 0 # initialize the dic; nothing is perturbed at the beginning

        for i in range(0, y_train.shape[0]):
            for eachelement in labeldic:
                if np.array_equal(Y_changed[i], labeldic[eachelement]) and (dic_label_newnum[eachelement] < float(dic_label_num[eachelement]) * noisedegree):
                    try:
                        Y_changed[i] = labeldic[eachelement + 1]
                        dic_label_newnum[eachelement] += 1
                    except:
                        Y_changed[i] = labeldic[0]
                        dic_label_newnum[eachelement] += 1
                    break
                else:
                    continue

        model.fit(X_train, Y_changed,batch_size=128,epochs=epoch, verbose=1) #retrain the model with perturbed labels

        trainaccuracy_perturbed = model.evaluate(X_train,y_train,verbose=0)[1]
        trainingacculist.append(trainaccuracy_perturbed)

        # get test accuracy when noise degree is zero
        if not testflag:
            testaccuracy = model.evaluate(X_hold, y_hold, verbose=0)[1]
            testflag = True


        cnt+=1

    Ytest = trainingacculist

    Xtest = noisedegreelist
    # m, b = poly.polyfit(Xtest, Ytest, 1) # conduct linear regression; b is the coefficient
    # pv = 1-abs(1-abs(b)) # mirror PV by one, so that PV increases up to 1, then worsen afterwards
    pv = trainaccuracy_perturbed

    print('PV: ' + str(round(pv, 2)))
    print('training accuracy: ' + str(round(trainingacculist[0], 2)))
    print('validation accuracy: ' + str(round(cvaccuracy, 2)))
    print('test accuracy: ' + str(round(testaccuracy, 2)))
    print('----------')

    # return dic
    dic_metric_value = {}
    dic_metric_value['pv'] = pv
    dic_metric_value['train'] = trainingacculist[0]
    dic_metric_value['test'] = testaccuracy
    dic_metric_value['cv'] = cvaccuracy

    return dic_metric_value


def get_noisydata_labelperturbation(y, noisedegree):
    '''
    Description
        This function perturbs noisedegree ratio of labels in the training data and return noisy data

    Parameters
        y: training data labels; type: numpy.ndarray
        noisedegree: the ratio of labels per class that are going to be perturbed

    Returns
        y_noisy: perturbed training data
    '''
    label_list = list(set(list(y)))

    dic_label_num = {}  # get the number of instances for each class

    for label in label_list:
        dic_label_num[label] = (y == label).sum()

    Y_changed = np.copy(y)
    dic_label_newnum = {}  # record the number of perturbed samples for each class

    for label in label_list:
        dic_label_newnum[label] = 0  # initialize the dic; nothing is perturbed at the beginning

    for i in range(0, Y_changed.size):
        cnnn = 0
        for label in label_list:
            # perturb the label only if the perturbed labels in this class are fewer than required,
            if Y_changed[i] == label_list[cnnn] and dic_label_newnum[label_list[cnnn]] < float(
                    dic_label_num[label_list[cnnn]]) * noisedegree:
                try:
                    # replace the current label with its right neighbour label in label_list
                    Y_changed[i] = label_list[cnnn + 1]
                    dic_label_newnum[label_list[cnnn]] += 1
                except:
                    # if the label is the last in the label_list, replace it with the first element in label_list
                    Y_changed[i] = label_list[0]
                    dic_label_newnum[label_list[cnnn]] += 1
                continue
            cnnn += 1

    return Y_changed

def get_slope_learningrate(filepath):
    lines = open(filepath).readlines()
    trainzerolist = []
    trainnoisylist = []

    cvlist = []
    testlist = []
    pvlist = []
    for line in lines:
        if 'algorithm' not in line:
            splits = line.split(',')
            if float(splits[1])==0.0:
                trainzerolist.append(splits[2])
                cvlist.append(float(splits[3]))
                testlist.append(float(splits[4]))
            if splits[1] == '0.2':
                trainnoisylist.append(splits[2])


    for i in np.arange(0,len(trainnoisylist),1):
        pvlist.append((float(trainzerolist[i])-float(trainnoisylist[i]))/0.20)

    return pvlist,cvlist,testlist










